#include "example.h"
#include <stdio.h>
#include <math.h>
#include <dlfcn.h>
#include <inttypes.h>

TG39DDC::TG39DDC(void)
{
    m_Device=NULL;
    m_API=NULL;
}

TG39DDC::~TG39DDC(void)
{
    Close();
    UnloadAPI();
}

bool TG39DDC::LoadAPI(void)
{
    if(m_API)
    {
        return true;
    }

    m_API=dlopen("libg39ddcapi.so",RTLD_LAZY);

    if(m_API==NULL)
    {
        printf("Failed to load libg39ddcapi.so. %s\n",dlerror());
        return false;
    }

    m_CreateInstance=(G39DDCAPI_CREATE_INSTANCE)dlsym(m_API,"CreateInstance");

    if(m_CreateInstance==NULL)
    {
        printf("Failed to retrieve address of the CreateInstance function!\n");
        dlclose(m_API);
        m_API=NULL;
        return false;
    }

    return true;
}

void TG39DDC::UnloadAPI(void)
{
    //release device object interface before the API is freed
    if(m_Device)
    {
        m_Device->Release();
        m_Device=NULL;
    }

    if(m_API)
    {
        dlclose(m_API);
        m_API=NULL;
    }
}

bool TG39DDC::Open(const char *DevicePath)
{
 G39DDC_DDC_INFO DDCInfo;
 uint32_t DDCTypeIndex;
 uint64_t Frequency;
 uint64_t FrontEndFrequency;
 int32_t DDC1Frequency;
 int32_t DDC2Frequency;
 int32_t DemodulatorFrequency;
 
    if(m_Device)
    {
        //close it if it is already open, if it is not open, the Close method does nothing
        m_Device->Close();
    }
    else
    {
        //device object has not been created yet
        if(!m_CreateInstance(G39DDC_CLASS_ID_DEVICE,(void**)&m_Device))
        {
            printf("Failed to create device object!\n");
            return false;
        }
    }

    if(!m_Device->Open(DevicePath))
    {
        printf("Failed to open device!\n");
        return false;
    }

    //power on
    m_Device->SetPower(1);

    //set DDC type of DDC1 (channel 0) to 12 (DDC Bandwidth=320 kHz, SampleRate=400 kHz, BitsPerSample=32)
    //number of available DDC types of DDC1 can be obtained using GetDDC1Count method
    m_Device->SetDDC1(0,12);

    //retrieve current DDC type of the DDC1 for channel 0
    m_Device->GetDDC1(0,&DDCTypeIndex);
    m_Device->GetDDCInfo(DDCTypeIndex,&DDCInfo);
    printf("DDC type of DDC1: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);


    //set DDC type of DDC2 (channel 0) to 12 (DDC Bandwidth=320 kHz, SampleRate=400 kHz, BitsPerSample=32)
    //number of available DDC types of DDC2 (for current DDC type of DDC1) can be obtained using GetDDC2Count method
    m_Device->SetDDC2(0,12);

    //retrieve current DDC type of the DDC2 for channel 0
    m_Device->GetDDC2(0,&DDCTypeIndex);
    m_Device->GetDDCInfo(DDCTypeIndex,&DDCInfo);
    printf("DDC type of DDC2: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);


    //set output audio sample rate index to 1 (48000 Hz)
    //number of available audio sample rates can be obtained using GetAudioSampleRateCount
    m_Device->SetAudio(0,1);

    //register callback object to this, the TG39DDC class implements methods of the IG39DDCDeviceCallback callback interface 
    m_Device->SetCallback(this);

    //tune channel 0 to an FM broadcasting station frequency
    m_Device->SetFrequency(0,87700000);
    //it is possible to tune front-end, DDC1, DDC2 and demodulator directly using
    //SetFrontEndFrequency, SetDDC1Frequency, SetDDC2Frequency and SetDemodulatorFrequency
    //SetFrequency(hDevice,0,87700000) can be replaced by the following:
    //SetFrontEndFrequency(90000000);
    //SetDDC1Frequency(0,-2300000);
    //SetDDC2Frequency(0,0);
    //SetDemodulatorFrequency(0,0);
    //final absolute frequency of the demodulator is 90000000-2300000+0+0=87700000

    //get absolute frequency where the demodulator of channel 0 is tuned
    m_Device->GetFrequency(0,&Frequency);

    //get front-end frequency
    m_Device->GetFrontEndFrequency(&FrontEndFrequency);

    //get DDC1 frequency
    m_Device->GetDDC1Frequency(0,&DDC1Frequency);

    //get relative DDC2 frequency for channel 0
    m_Device->GetDDC2Frequency(0,&DDC2Frequency);

    //get relative demodulator frequency for channel 0
    m_Device->GetDemodulatorFrequency(0,&DemodulatorFrequency);

    printf("Listening to (demodulator absolute frequency): %" PRIu64 " Hz\n",Frequency);
    printf("Front-end frequency: %"PRIu64" Hz\n",FrontEndFrequency);
    printf("DDC1 frequency (relative): %d Hz\n",DDC1Frequency);
    printf("DDC2 frequency (relative): %d Hz\n",DDC2Frequency);
    printf("Demodulator frequency (relative): %d Hz\n",DemodulatorFrequency);


    //disable AGC for channel 0
    m_Device->SetAGC(0,0);

    //set FMW mode for channel 0
    m_Device->SetDemodulatorMode(0,G39DDC_MODE_FMW);

    //set demodulator filter bandwidth for channel 0 to 250kHz
    m_Device->SetDemodulatorFilterBandwidth(0,250000);

    return true;
}

void TG39DDC::StartStreaming(void)
{
 uint32_t SamplesPerBuffer;
 G39DDC_DDC_INFO DDCInfo;
 uint32_t DDCTypeIndex;
 uint32_t AudioSampleRateIndex;
 uint32_t AudioSampleRate;

    if(m_Device)
    {
        //start DDC1 streaming using ~50ms long buffers
        m_Device->GetDDC1(0,&DDCTypeIndex);             //get current DDC type index of DDC1 for channel 0
        m_Device->GetDDCInfo(DDCTypeIndex,&DDCInfo);    //get DDC info for DDC type index
        SamplesPerBuffer=50*DDCInfo.SampleRate/1000;    //calculte number of I/Q sample-sets per 50ms buffer
        SamplesPerBuffer=(SamplesPerBuffer+63)&(~63);   //round up to 64
        m_Device->StartDDC1(0,SamplesPerBuffer);        //start DDC1 streaming

        //start DDC2 streaming using ~50ms long buffers
        m_Device->GetDDC2(0,&DDCTypeIndex);             //get current DDC type index of DDC2 for channel 0
        m_Device->GetDDCInfo(DDCTypeIndex,&DDCInfo);    //get DDC info for DDC type index
        SamplesPerBuffer=50*DDCInfo.SampleRate/1000;    //calculte number of I/Q sample-sets per 50ms buffer
        SamplesPerBuffer=(SamplesPerBuffer+63)&(~63);   //round up to 64
        m_Device->StartDDC2(0,SamplesPerBuffer);        //start DDC2 streaming

        //start audio streaming using ~50ms long buffers
        m_Device->GetAudio(0,&AudioSampleRateIndex);    //get current audio sampel rate index
        m_Device->GetAudioSampleRate(AudioSampleRateIndex,&AudioSampleRate);  //get audio sample rate by its index
        SamplesPerBuffer=50*AudioSampleRate/1000;       //calculte number of stereo audio sample-sets per 50ms buffer
        SamplesPerBuffer=(SamplesPerBuffer+63)&(~63);   //round up to 64
        m_Device->StartAudio(0,SamplesPerBuffer);       //start audio streaming
    }
}

void TG39DDC::Close(void)
{
    if(m_Device)
    {
        //stop streaming if it is running
        StopStreaming();

        //close wave out if it is open
        CloseWaveOut();

        //close the receiver if it is open
        m_Device->Close();
    }
}

void TG39DDC::StopStreaming(void)
{
    if(m_Device)
    {
        //stop audio, DDC2 and DDC1 streaming for channel 0
        m_Device->StopAudio(0);
        m_Device->StopDDC2(0);
        m_Device->StopDDC1(0);
    }
}

void TG39DDC::OpenWaveOut(void)
{ 
 uint32_t AudioSampleRateIndex,AudioSampleRate;
 pa_sample_spec SampleFormat;
 int Error;
 
    if(m_PaSimple || m_Device==NULL)
    {
        return;
    }
        
    //retrieve current audio sample rate index of channel 0
    m_Device->GetAudio(0,&AudioSampleRateIndex);
    m_Device->GetAudioSampleRate(AudioSampleRateIndex,&AudioSampleRate);

    //initialize pulse audio
    SampleFormat.channels=2;
    SampleFormat.rate=AudioSampleRate;
    SampleFormat.format=PA_SAMPLE_FLOAT32LE;
    
    m_PaSimple=pa_simple_new(NULL,"G39DDC SDK EXAMPLE",PA_STREAM_PLAYBACK,NULL,"Audio0",&SampleFormat,NULL,NULL,&Error);
    
    if(m_PaSimple==NULL)
    {
        printf("Failed to initialize pulse audio.\n");
    }
}

void TG39DDC::CloseWaveOut(void)
{
    if(m_PaSimple)
    {
        pa_simple_free(m_PaSimple);
        m_PaSimple=NULL;
    }
}


void TG39DDC::G39DDC_DDC1StreamCallback(IG39DDCDevice *Device,uint32_t Channel,const void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample)
{
    //it does nothing in this example
}

int TG39DDC::G39DDC_DDC1PlaybackStreamCallback(IG39DDCDevice *Device,uint32_t Channel,void *Buffer,uint32_t NumberOfSamples,uint32_t BitsPerSample)
{
    //it does nothing in this example
    return 0;
}

void TG39DDC::G39DDC_DDC2StreamCallback(IG39DDCDevice *Device,uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples)
{
    //it does nothing in this example
}

void TG39DDC::G39DDC_DDC2PreprocessedStreamCallback(IG39DDCDevice *Device,uint32_t Channel,const float *Buffer,uint32_t NumberOfSamples,float SlevelPeak,float SlevelRMS)
{
 double Slevel_dBm;

    if(Channel==0)
    {
        //convert slevel RMS in Volts to dBm
        Slevel_dBm=10.0*log10(SlevelRMS*SlevelRMS*(1000.0/50.0));

        printf("\rSlevel: RMS[V]=%9.7f V, RMS[dBm]=%6.1f dBm, Peak[V]=%9.7f V",SlevelRMS,Slevel_dBm,SlevelPeak);
        fflush(stdout);
    }
}

void TG39DDC::G39DDC_IFCallback(IG39DDCDevice *Device,const int16_t *Buffer,uint32_t NumberOfSamples,uint32_t CenterFrequency,uint16_t MaxADCAmplitude,uint32_t ADCSampleRate)
{
    //it does nothing in this example
}

int TG39DDC::G39DDC_SweepingCallback(IG39DDCDevice *Device,const int16_t *Buffer,uint32_t NumberOfSamples,uint32_t CenterFrequency,uint64_t FrontEndFrequency)
{
    //it does nothing in this example
    return 0;
}

void TG39DDC::G39DDC_AudioStreamCallback(IG39DDCDevice *Device,uint32_t Channel,uint32_t Type,const float *Buffer,uint32_t NumberOfSamples)
{
 int Error;

    if(Type==2 && m_PaSimple!=NULL)
    {
        //we want to play only audio buffers of type 2 which means final audio samples (audio gain, audio filter and volume are applied)
        pa_simple_write(m_PaSimple,Buffer,NumberOfSamples*2*sizeof(float),&Error);
    } 
}

int TG39DDC::G39DDC_AudioPlaybackStreamCallback(IG39DDCDevice *Device,uint32_t Channel,float *Buffer,uint32_t NumberOfSamples)
{
    //it does nothing in this example
    return 0;
}


int main(void)
{
 TG39DDC g39ddc;

    //load G39DDC API
    if(!g39ddc.LoadAPI())
    {
        return -1;
    }

    //open the first availbe G39DDC receiver and setup it
    if(!g39ddc.OpenFirst())
    {
        return -1;
    }
    
    g39ddc.OpenWaveOut();
    
    printf("\nPress ENTER to exit\n\n");

    //start streaming
    g39ddc.StartStreaming();

    //wait for ENTER
    getchar();

    //stop streaming
    g39ddc.StopStreaming();    

    //close the G39DDC receiver
    g39ddc.Close();
    
    g39ddc.CloseWaveOut();

    //unload G39DDC API
    g39ddc.UnloadAPI();
    
    return 0;
}